/* 2018-05 read tridiag-file and compute eigenvalues (for FTLM)
   usefull if spinpack has aborted near end to get eigenvalues
   txt-format: idx a[i] b[i] time, dat-format: a[i] b[i]
   - tri.txt = (3+21+21+8+1=54)chars/line*lines75=4050
   - tri.dat = (8+8)Byte/line binary double
   see spinstri.c + diag.c CFG_FTLM
   
  CentOS7: yum install openblas
  compile: cc [-DCFG_FTLM=1 [-DLAPACK -lopenblas]] -lm -o tridiag \
           ../utils/tridiag.c
  running: ./tridiag [tmp/tri.dat]   # optional filename (.dat or .txt)
  example1: gcc -DCFG_FTLM=1 -o tridiag -lm ../utils/tridiag.c
            ./tridiag
  example2: icc -o tridiag ../utils/tridiag.c -DCFG_FTLM=1 -DLAPACK \
             -lm -g ${MKL_SHLIB}
   
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>     /* fabs, sqrt */
#include <string.h>  /* strlen */

#define MAX_ITR 1024
static double ta[MAX_ITR], tb[MAX_ITR]; /* ta,tb Tridiagonalmatrix */
/* fortran like 2D = bad */
/* NEV is MAX_ITR for CFG_FTLM */
#define NEV MAX_ITR
static double ew[NEV]; /* eigenvalues */
static double tc[MAX_ITR][NEV]; /* eigenvectors of tridiagonalmatrix    */

#define Norm(x) norm(x)   /* r*r+i*i */
#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))
static inline double norm(const double r){return(r*r);}
/* ----------- bisection method -------------------------------------
 * calculate the eigenvalues of a tridiagonal matrix T=(tb,ta,tb)
 *  Inp: ta,tb=Tridiagonalmatrix(T),  nn=size of the matrix T
 *       ne=number of eigenvalues ew to calculate, ep=precision
 *  Out: ew  eigenvalues of T via "Nullstellen" of Det(A-x*I)
 *        ToDo: *ew as argument!!!
 *  What happens for degeneracy?
 *    => EW only one-fold, but should be n-fold (as expected by vec12)
 *    ToDo: after EW[ne] is converged, call another exact/slow tridiag
 *          p.e. dsterf() of lapack/icepack ? t(300)=126s acceptable
 *          sparc-Ultra1-Boson: time/s = 1.5e-6 dim ^ 3.2 
 *      OR try (modified) vec12 morefold on same eigenvalue until <tc>=0
 *      check on til8_18b.dat
 * add ta[1..nn-1],tb[0..nn-2] as parameter 2012-05
 * --- */
static void bisec(const int nn,const int ne, /* input */
    const double ep,        /* input rel. precision 1e-13 (max double ...) */
    const double *ta, const double *tb, /* input tridiag matrix ta[nn], tb[nn-1] */
    double *ew){            /* output lowest eigenvalues ew[ne] */
  static double b2[MAX_ITR]; /* if not enough stackspace, faster */
  double ra,rb,rc,g,epa,range; int i,j,k,neg,ipass;
  /* find a interval containing all the eigenvalues rb..ra */
  range=fabs(ta[0])+fabs(tb[0]);
  for(k=1;k<nn-1;k++) range=max(range,fabs(tb[k-1])+fabs(ta[k])+fabs(tb[k]));
  range=max(range,fabs(tb[nn-2])+fabs(ta[nn-1])); range=-range;
  /* range = max_i (sum_j fabs(matrix_ij))  matrix line */
  b2[0]=0; for(i=1;i<nn;i++) b2[i]=Norm(tb[i-1]);
  epa=fabs(range*ep); for(i=0;i<ne;i++) ew[i]=-range;
  rb=range;
  /* find the eigenvalues of a tridiagonal matrix * 
   * lying in a given interval rb...ra            */
  for(k=0;k<ne;k++) { ra=ew[k];       /* ra=upper bound, rb=lower bound */
    /* was j<100, but this is a bit to small ??? */
    for(j=0;j<10000;j++){		/* sturm sequencing */
      if(fabs(ra-rb)<epa)break;	/* ra=rb => break */
      rc=(ra+rb)*0.5; ipass=neg=0; g=1.0;
      for(i=0;i<nn;i++){   /* g=Det_{i}(rc)/Det_{i-1}(rc) */
        if(ipass==0)g=rc-ta[i]-b2[i]/g;else	/* do not divide if g=0 */
        if(ipass==1)ipass=2;
        else      { ipass=0; g=rc-ta[i]; }	/* new start, decoupled */
        if(ipass==0){ if(g<0.0)neg++;		/* sign-change of Det */
          if(fabs(g)<=fabs(b2[i]*epa*ep)) ipass=1; } /* g==0 ? */
      }   
      neg=nn-neg;   /* num of nullstellen rc...ra -> rb...rc */
      if(neg<=k)rb=rc;		/* new lower bound */
      else    { ra=rc; 		/* new upper bound */
                for(i=k;i<min(neg,ne);i++)ew[i]=rc; }
    } }
  return; }

/* --- eigenvectors (EV) of a tridiagonal matrix by inverse iteration
 *  http://en.wikipedia.org/wiki/Power_iteration Mai2010
 *     inverse iteration == power iteration for inverse matrix (A^(-1))
 *                          + choice to choose the eigenvalue to converge to
 *  see Numerical Recipes  (Feb99)
 *   A - matrix
 *   a - start-eigenvalue
 *   x - startvector (random)
 *   step1: solve equation system (A-a*1)*y==x for vector y (ex: LU decomposition)
 *   step2: x := y/|y|          (normalize)
 *   step3: a := a+(x,x)/(x,y)  (not necessary every step)
 *   if Ax=ax then break else goto step1   
 *
 * --- nn i matrix size, {ta,tb} TriDiagMatrix -> tc EVs
 * --- for degeneracy EW must be n-fold in ew[] and nev>=n!  (does not work!)
 *   ex: ew[]={ -3.5, -3.5, -3.5, -1.5} for 3-fold degeneracy
 *   input: nn - number of sites, nev - number eigenvectors,
 *          ne0 - offset for lowest eigenvectors (0..n1-NEV)
 * ToDo: degenerated test-system?
 */
void vec12(const int nn, const int nev, const int ne0,
     const double *ew, const double *ta, const double *tb, /* 2016-04 added */
     /* 2016-05-18 NEV must be at least MAX_ITR for CFG_FTLM */
     double (*tc)[NEV]      /* only lowest eigenstates wanted */
     ){
  static double di[MAX_ITR],bl[MAX_ITR],bu[MAX_ITR],
                bv[MAX_ITR],cm[MAX_ITR];
  static char  lex[MAX_ITR];	/* 0 (non pivoting) or 1 (pivoting) */
    double s,prd; int i,j,k,km,l;
    km=0;/* km=0 added by JS */
    /* 2015-12-07 stop array overflow warning, gcc-4.8 */
#ifndef CFG_FTLM  /* ~ Finite-Temperature Lanczos Method */
    if (nn> MAX_ITR || nev> NEV) F0("avoid array overflow");
    if (nn<=MAX_ITR && nev<=NEV)                            
#endif
    for(k=0;k<nev;k++){ /* if ew0==ew1==ew2 : 3 orthogonal vectors */
        for(j=0;j<nn;j++) {di[j]=ew[k+ne0]-ta[j]; bl[j]=bu[j]=-tb[j];}
        /* --- LU decomposition, prepare lex, cm, bu, bv, di */ 
        for(j=0;j<nn-1;j++) {
           if(fabs(di[j])>fabs(bl[j])){		/* --- non pivoting */
             lex[j]=0; if(fabs(di[j])<1e-30) di[j]=1e-30; /* e-40 => e-30 (float) */
             cm[j+1]=bl[j]/di[j]; di[j+1]-=cm[j+1]*bu[j]; bv[j]=0; }  
           else {				/* --- pivoting */
             lex[j]=1; cm[j+1]=di[j]/bl[j]; di[j]=bl[j]; s=bu[j];
             bu[j]=di[j+1]; bv[j]=bu[j+1]; di[j+1]=s-cm[j+1]*bu[j];
             bu[j+1]=-cm[j+1]*bv[j]; }
        }  
        if(fabs(di[nn-1])<1e-30) di[nn-1]=1e-30;
        /* --- initial vec tc */ 
        for(j=0;j<nn;j++) tc[j][k]=1./((j+1)*5.); /* old code, random ??? */
//      tc[k%nn][k]=-0.5; /* 2016-05-04 add different startvectors! FTLM */
        /* --- degeneracy check up -> Entartung km..k */
        /*  degeneracy is given in ew[] implicitely!  */
        if(k==0) km=k; else   /* ToDo: FTLM-problem 2016 with degeneracy ? */
        if(fabs(ew[k+ne0]-ew[km+ne0])>1e-10) km=k;     /* old=1e-13 1998 */ 
        else for(i=km;i<k;i++){       /* orthogonalise ini_vec to tri_ev */
          /* ToDo: exchange array indexing for better access pattern? */
          /* gcc warns about array bounds on 2nd line (not understandable) */
          for(prd=0,j=0;j<nn;j++) prd+=tc[j][i]*tc[j][k];  /* <c_i|c_k> */
          for(      j=0;j<nn;j++) tc[j][k]-=prd*tc[j][i]; }
        /* ToDo: stop and ignore EW, if <tc|tc>==0 ? */
        /* --- inverse iteration of tc (inp: lex, cm, bu, bv, di) */
        for(l=0;l<(k-km+3);l++){	/* what does l??? */
          if((l!=0)||(k!=km)){ /* --- forward substitution */
            for(j=0;j<nn-1;j++)
            if(lex[j]==0) tc[j+1][k]-=cm[j+1]*tc[j][k];
            else {s=tc[j][k];tc[j][k]=tc[j+1][k];
                  tc[j+1][k]=s-cm[j+1]*tc[j][k];}
          }  /* --- backward substitution */
          for(j=nn-1;j>=0;j--){
            s=tc[j][k];
            if(j<(nn-1))s-=bu[j]*tc[j+1][k];
            if(j<(nn-2))s-=bv[j]*tc[j+2][k];
            tc[j][k]=s/di[j];
          }   /* --- normalization */
          for(prd=0,j=0;j<nn;j++) prd+=Norm(tc[j][k]);
          if(prd>1e-30)prd=1./sqrt(prd);else prd=0;
          for(j=0;j<nn;j++) tc[j][k]*=prd;
        } /* for l */
    } /* for k=0..nev-1 */
    return;
}



int main(int argn, char *argv[]){
  char *fname = "tmp/tri.dat";  /* filename tridiag-values */
  static FILE *f1=NULL;
  char s2[256]; /* input buffer txt-file */
  int i1, l1 /* length fname */, idx=0, itr=0 /* idx read */;
  if (argn>1 && argv[1]) fname = argv[1]; /* fname argument */
  l1=strlen(fname);
  printf("DBG: fname = \"%s\"  len=%d\n", fname, l1);
  for (i1=0;i1<MAX_ITR;i1++) ta[i1]=tb[i1]=0; // ini
  // file name must end with .dat or .txt
  if (fname[l1-4]!='.') { printf("ERROR: expect .dat or .txt\n"); exit(1); }
  if (fname[l1-3]=='d') {
    f1= fopen(fname,"rb");
    if (!f1) { printf("ERR: open file\n");exit(2); }
    while ((!feof(f1)) && itr<MAX_ITR) { // unsure
      if (fread(ta+itr,sizeof(double),1,f1)!=1) break;
      if (fread(tb+itr,sizeof(double),1,f1)!=1) break;
      itr++;
    }
  } else if (fname[l1-3]=='t') {
    f1= fopen(fname,"ra");
    if (!f1) { printf("ERR: open file\n");exit(2); }
    while ((!feof(f1)) && itr<MAX_ITR) { // unsure
      if (fgets(s2,256,f1)==NULL) break; // read max 255 chars incl. \n
      if (sscanf(s2,"%d %lf %lf %*s",&idx,ta+itr,tb+itr)!=3) {
        printf("ERR: read txt L%d: %s\n",itr,s2); break; }
      itr++;
    }
  } else { printf("ERROR: expect .dat or .txt\n"); exit(1); }
  printf("... read %d lines\n",itr);
  if (f1) fclose(f1);
#ifdef LAPACK
  { int i1, i2, info; static double work[MAX_ITR*MAX_ITR]; /* 8MB on stack!? */
    static double a[MAX_ITR], b[MAX_ITR], c[MAX_ITR*MAX_ITR]; /* 8MB */
    for (i1=0;i1<itr;i1++) a[i1]=ta[i1];
    for (i1=0;i1<itr;i1++) b[i1]=tb[i1];
    for (i1=0;i1<itr;i1++) /* tc must be initialized as identity-matrix 1 */
    for (i2=0;i2<itr;i2++) c[i1*itr+i2]=0.0;
    for (i1=0;i1<itr;i1++) c[i1*itr+i1]=1.0;
    dsteqr_("I",&itr,a,b,c,&itr,work,&info); /* destroy a,b,c=tc[0] */
    { int i1; double s2=0.0, trace=0.0;
      for (i1=0;i1<itr;i1++) {
        printf("\nFTLM: i,ew_i,tri_ev_i[0]= %3d %16.12f %16.12f dsteqr",
               i1, a[i1],   c[i1*itr+0]);
        s2 += c[i1*itr+0] * c[i1*itr+0]; 
        trace += a[i1];
      } /* i1-loop */  
      printf("\nFTLM: sum_E,tri_ev_i[0]^2= %19.12f %16.12f", trace, s2);
    } /* !mpi_i */
  }
#endif
 bisec(itr,/* NEW*/itr,1e-14/*1e-13*/,ta,tb,ew);  /* out to ew[NEW] */
 /* ToDo: missing iteration? last two vectors have deviations! */
 vec12(itr,/*nev*/itr,0,ew,ta,tb,tc); /* get tridiagonal eigenvectors tc[0..i][
 /* vec12() gives bad tc-results on degenerated eigenvalues */
 /* kago36 nud=32+4 Sz=14 k=-999 8fold, but dsteqr/bisec dont show 8*E=6.0
  *   problem starts with i=213 3*17.5 than 4*18.0 ToDo: why never get multiple
  */
 { int i1; double s2=0.0, trace=0.0;
   for (i1=0;i1<itr;i1++) {
     printf("\nFTLM: i,ew_i,tri_ev_i[0]= %3d %16.12f %16.12f bisec(1e-14)",
           i1, ew[i1], tc[0][i1]);
     s2 += tc[0][i1] * tc[0][i1]; 
     trace += ew[i1];
   } /* i1-loop */   
   printf("\nFTLM: sum_E,tri_ev_i[0]^2= %19.12f %16.12f", trace, s2);
   if (fabs(s2-1.)>1e-6) 
     printf("\nWARNING: sum_ev0!=1, bad bisec for degenerated ew");
 } /* !mpi_i */
 printf("\n-END-\n");
} 

